from ._col_gen_classifier import BaseMasterProblem
from ._col_gen_classifier import BaseSubproblem
from ._col_gen_classifier import ColGenClassifier
from ._bdr_classifier import BDRMasterProblem
from ._bdr_classifier import BDRSubProblem
from ._bdr_classifier import BDRHeuristic
from ._bdr_classifier import BooleanDecisionRuleClassifier
from ._bdr_classifier import BooleanDecisionRuleClassifierWithHeuristic
from ._dtree_classifier import Path, Node, Split, Leaf
from ._dtree_classifier import DTreeMasterProblem, DTreeSubProblem
from ._dtree_classifier import DTreeClassifier

from ._version import __version__

__all__ = ['BaseMasterProblem', 'BaseSubproblem', 'ColGenClassifier',
           'BDRMasterProblem', 'BDRSubProblem', 'BDRHeuristic',
           'BooleanDecisionRuleClassifier',
           'BooleanDecisionRuleClassifierWithHeuristic',
           'Path', 'Node', 'Split', 'Leaf',
           'DTreeMasterProblem', 'DTreeSubProblem',
           'DTreeClassifier',
           '__version__']



import pytest

from sklearn.utils.estimator_checks import check_estimator

from col_gen_estimator import BooleanDecisionRuleClassifier
from col_gen_estimator import BooleanDecisionRuleClassifierWithHeuristic
from col_gen_estimator import DTreeClassifier


@pytest.mark.parametrize(
    "estimator",
    [BooleanDecisionRuleClassifier(),
     BooleanDecisionRuleClassifierWithHeuristic(),
     DTreeClassifier()]
)
def test_all_estimators(estimator):
    return check_estimator(estimator)



import pytest
import numpy as np

from numpy.testing import assert_array_equal
from numpy.testing import assert_equal
from numpy.testing import assert_raises

from col_gen_estimator import DTreeMasterProblem
from col_gen_estimator import DTreeSubProblem
from col_gen_estimator import DTreeClassifier
from col_gen_estimator import Split
from col_gen_estimator import Node
from col_gen_estimator import Leaf
from col_gen_estimator import Path


def test_leaf_creation():
    leaf = Leaf()
    leaf.create_leaf(0, 2)
    assert_array_equal(leaf.left_nodes, [1, 0])
    assert_array_equal(leaf.right_nodes, [])

    leaf.create_leaf(1, 2)
    assert_array_equal(leaf.left_nodes, [0])
    assert_array_equal(leaf.right_nodes, [1])

    leaf.create_leaf(2, 2)
    assert_array_equal(leaf.left_nodes, [2])
    assert_array_equal(leaf.right_nodes, [0])

    leaf.create_leaf(3, 2)
    assert_array_equal(leaf.left_nodes, [])
    assert_array_equal(leaf.right_nodes, [2, 0])


def test_path_set_leaf():
    path = Path()
    path.set_leaf(0, 2)
    assert_array_equal(path.node_ids, [1, 0])
    assert(path.leaf_id, 0)

    path.set_leaf(1, 2)
    assert_array_equal(path.node_ids, [1, 0])
    assert(path.leaf_id, 1)

    path.set_leaf(2, 2)
    assert_array_equal(path.node_ids, [2, 0])
    assert(path.leaf_id, 2)

    path.set_leaf(3, 2)
    assert_array_equal(path.node_ids, [2, 0])
    assert(path.leaf_id, 3)


@pytest.fixture
def data():
    X = np.array([[1, 1, 1],
                  [1, 1, 0],
                  [1, 0, 1],
                  [1, 0, 0],
                  [0, 1, 1],
                  [0, 1, 0],
                  [0, 0, 1],
                  [0, 0, 0], ], np.int8)
    y = np.array([1, 1, 0, 0, 0, 1, 0, 1], np.int8)

    splitA = Split()
    splitA.id = 0
    splitA.feature = 0
    splitA.threshold = 0.5

    splitB = Split()
    splitB.id = 1
    splitB.feature = 1
    splitB.threshold = 0.5

    splitC = Split()
    splitC.id = 2
    splitC.feature = 2
    splitC.threshold = 0.5

    root_node = Node()
    root_node.id = 0
    root_node.candidate_splits = [0]

    left_node = Node()
    left_node.id = 1
    left_node.candidate_splits = [1, 2]

    right_node = Node()
    right_node.id = 2
    right_node.candidate_splits = [1, 2]

    leaf_0 = Leaf()
    leaf_0.create_leaf(0, 2)

    leaf_1 = Leaf()
    leaf_1.create_leaf(1, 2)

    leaf_2 = Leaf()
    leaf_2.create_leaf(2, 2)

    leaf_3 = Leaf()
    leaf_3.create_leaf(3, 2)

    path_0 = Path()
    path_0.set_leaf(0, 2)
    path_0.splits = [1, 0]
    path_0.cost = 1
    path_0.target = 1

    path_1 = Path()
    path_1.set_leaf(1, 2)
    path_1.splits = [1, 0]
    path_1.cost = 1
    path_1.target = 0

    path_2 = Path()
    path_2.set_leaf(2, 2)
    path_2.splits = [2, 0]
    path_2.cost = 1
    path_2.target = 0

    path_3 = Path()
    path_3.set_leaf(3, 2)
    path_3.splits = [2, 0]
    path_3.cost = 1
    path_3.target = 1

    path_4 = Path()
    path_4.set_leaf(0, 2)
    path_4.splits = [2, 0]
    path_4.cost = 2
    path_4.target = 1

    path_5 = Path()
    path_5.set_leaf(1, 2)
    path_5.splits = [2, 0]
    path_5.cost = 2
    path_5.target = 0

    path_6 = Path()
    path_6.set_leaf(2, 2)
    path_6.splits = [1, 0]
    path_6.cost = 2
    path_6.target = 0

    path_7 = Path()
    path_7.set_leaf(3, 2)
    path_7.splits = [1, 0]
    path_7.cost = 2
    path_7.target = 1

    paths = [path_0, path_1, path_2, path_3, path_4, path_5, path_6, path_7]
    nodes = [root_node, left_node, right_node]
    leaves = [leaf_0, leaf_1, leaf_2, leaf_3]
    splits = [splitA, splitB, splitC]
    targets = [0, 1]

    return (X, y, paths, nodes, leaves, splits, targets)


def test_master_prob(data):
    X = data[0]
    y = data[1]
    paths = data[2]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    master_problem = DTreeMasterProblem(paths, leaves, nodes, splits)
    assert_equal(master_problem.generated_, False)
    master_problem.generate_mp(X, y)
    assert_equal(master_problem.generated_, True)
    duals = master_problem.solve_rmp()
    assert(len(duals), 3)
    leaf_duals = duals[0]
    row_duals = duals[1]
    ns_duals = duals[2]
    assert_array_equal(leaf_duals, [1, 0, 2, 1])
    assert_array_equal(row_duals, [0, -1, 0, 0, 0, 1, 0, 0])
    assert_equal(ns_duals[1][1][2], 2)
    assert_equal(ns_duals[3][2][1], 2)


def test_master_prob_add_col(data):
    X = data[0]
    y = data[1]
    paths = data[2][:7]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    master_problem = DTreeMasterProblem(paths, leaves, nodes, splits)
    assert_equal(master_problem.generated_, False)
    master_problem.generate_mp(X, y)
    assert_equal(master_problem.generated_, True)
    duals = master_problem.solve_rmp()
    assert(len(duals), 3)
    leaf_duals = duals[0]
    row_duals = duals[1]
    ns_duals = duals[2]
    assert_array_equal(leaf_duals, [1, 0, 0, 0])
    assert_array_equal(row_duals, [0, 0, 1, 0, 0, 1, 0, 0])
    assert_equal(ns_duals[1][1][2], 2)

    for leaf in leaves:
        for node in nodes:
            if node.id not in leaf.left_nodes + leaf.right_nodes:
                continue
            for split_id in node.candidate_splits:
                print("leaf={leaf}, node={node}, split={split},\
                      dual={dual}".format(
                    leaf=leaf.id, node=node.id, split=split_id,
                    dual=ns_duals[leaf.id][node.id][split_id]))
    path_7 = data[2][-1]
    master_problem.add_column(path_7)
    duals = master_problem.solve_rmp()
    assert(len(duals), 3)
    leaf_duals = duals[0]
    row_duals = duals[1]
    ns_duals = duals[2]
    assert_array_equal(leaf_duals, [1, 0, 2, 1])
    assert_array_equal(row_duals, [0, -1, 0, 0, 0, 1, 0, 0])
    assert_equal(ns_duals[1][1][2], 2)
    assert_equal(ns_duals[3][2][1], 2)


def test_subproblem(data):
    X = data[0]
    y = data[1]
    paths = data[2]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    targets = data[6]
    subproblem = DTreeSubProblem(leaves[0], nodes, splits, targets, depth=2)

    leaf_duals = [2, 3, 1, 2]
    row_duals = [0, 1, 0, 0, 0, -1, 0, 0]
    ns_duals = {}

    ns_duals[0] = {}
    zero_nodes = leaves[0].left_nodes + leaves[0].right_nodes
    for node in nodes:
        if node.id not in zero_nodes:
            continue
        ns_duals[0][node.id] = {}
        for split_id in node.candidate_splits:
            ns_duals[0][node.id][split_id] = 0.0
    duals = (leaf_duals, row_duals, ns_duals)
    new_paths = subproblem.generate_columns(X, y, duals)
    assert_equal(len(new_paths), 1)
    path = new_paths[0]
    assert_equal(path.leaf_id, 0)
    assert_array_equal(path.node_ids, [0, 1])
    assert_array_equal(path.splits, [0, 2])
    assert_equal(path.cost, 2)
    assert_equal(path.target, 1)


def test_subproblem2(data):
    X = data[0]
    y = data[1]
    paths = data[2]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    targets = data[6]
    subproblem = DTreeSubProblem(leaves[0], nodes, splits, targets, depth=2)

    leaf_duals = [1, 0, 2, 1]
    row_duals = [0, -1, 0, 0, 0, 1, 0, 0]
    ns_duals = {}

    ns_duals[0] = {}
    zero_nodes = leaves[0].left_nodes + leaves[0].right_nodes
    for node in nodes:
        if node.id not in zero_nodes:
            continue
        ns_duals[0][node.id] = {}
        for split_id in node.candidate_splits:
            ns_duals[0][node.id][split_id] = 0.0
    duals = (leaf_duals, row_duals, ns_duals)
    new_paths = subproblem.generate_columns(X, y, duals)
    assert_array_equal(new_paths, [])


def test_subproblem3(data):
    X = data[0]
    y = data[1]
    paths = data[2]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    targets = data[6]
    subproblem = DTreeSubProblem(leaves[2], nodes, splits, targets, depth=2)

    leaf_duals = [1, 0, 0, 0]
    row_duals = [0, 0, 1, 0, 0, 1, 0, 0]
    ns_duals = {}
    for node in nodes:
        print(node.candidate_splits)

    ns_duals[2] = {}
    two_nodes = leaves[2].left_nodes + leaves[2].right_nodes
    for node in nodes:
        if node.id not in two_nodes:
            continue
        ns_duals[2][node.id] = {}
        for split_id in node.candidate_splits:
            ns_duals[2][node.id][split_id] = 0.0
    ns_duals[2][0][0] = 1.0
    duals = (leaf_duals, row_duals, ns_duals)
    new_paths = subproblem.generate_columns(X, y, duals)
    assert_array_equal(new_paths, [])


def test_subproblem4(data):
    X = data[0]
    y = data[1]
    paths = data[2]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    targets = data[6]
    subproblem = DTreeSubProblem(leaves[3], nodes, splits, targets, depth=2)

    leaf_duals = [1, 0, 0, 0]
    row_duals = [0, 0, 1, 0, 0, 1, 0, 0]
    ns_duals = {}
    for node in nodes:
        print(node.candidate_splits)

    ns_duals[3] = {}
    two_nodes = leaves[3].left_nodes + leaves[3].right_nodes
    for node in nodes:
        if node.id not in two_nodes:
            continue
        ns_duals[3][node.id] = {}
        for split_id in node.candidate_splits:
            ns_duals[3][node.id][split_id] = 0.0
    duals = (leaf_duals, row_duals, ns_duals)
    new_paths = subproblem.generate_columns(X, y, duals)
    assert_equal(len(new_paths), 1)
    path = new_paths[0]
    assert_equal(path.leaf_id, 3)
    assert_array_equal(path.node_ids, [0, 2])
    assert_array_equal(path.splits, [0, 1])
    assert_equal(path.cost, 2)
    assert_equal(path.target, 1)


def test_fit_predict(data):
    X = data[0]
    y = data[1]
    paths = data[2][:7]
    nodes = data[3]
    leaves = data[4]
    splits = data[5]
    clf = DTreeClassifier(paths, leaves, nodes, splits,
                          tree_depth=2, targets=[0, 1], max_iterations=3)
    clf.fit(X, y)
    assert hasattr(clf, 'is_fitted_')
    assert hasattr(clf, 'classes_')
    assert hasattr(clf, 'X_')
    assert hasattr(clf, 'y_')
    assert hasattr(clf, 'mp_optimal_')
    assert hasattr(clf, 'performed_iter_')
    assert hasattr(clf, 'num_col_added_sp_')

    assert_equal(clf.mp_optimal_, True)
    assert_equal(clf.performed_iter_, 2)
    assert_equal(clf.num_col_added_sp_[0], 0)
    assert_equal(clf.num_col_added_sp_[1], 0)
    assert_equal(clf.num_col_added_sp_[2], 0)
    assert_equal(clf.num_col_added_sp_[3], 1)

    y_pred = clf.predict(X)
    assert y_pred.shape == (X.shape[0],)
    assert_array_equal(y_pred, y)



"""
Binary classifier using boolean decision rule generation method.

Reference: Sanjeeb Dash, Oktay Gunluk, and Dennis Wei. Boolean decision rules
via column generation. In Advances in Neural Information Processing Systems,
volume 31. Curran Associates, Inc., 2018.

The model is designed for binary classification. The model generates a boolean
decision rule in DNF (Disjunctive Normal Form, OR-of-ANDs). An example of a DNF
rule with two clauses is "(x1 AND x2) OR (x3 AND x5 AND x7)". These rules
return True when the example belongs to the class and False otherwise. These
binary rules also serve as explanation for the output class.

Sets
P set of all positive examples in the dataset.
Z set of all negative examples in the dataset.
K set of all possible clauses.
K_i set of all possible clauses satisfied by input i.

Parameters
c_k complexity of the clause k. Typically set to 1 + length of the clause.
C complexity of the explanation.
p misclassification cost weight for false negatives.

Decision Variables
w_k binary variable indicating if the clause k is selected.
xi_i binary variable indicating if the input i does not satisfy any selected
     clauses.

z_{MIP} = min   sum_{i in P} p*xi_i + sum_{i in Z} sum_{k in K_i} w_k   (obj)
          s.t.  xi_i + sum_{k in K_i} w_k >= 1, i in P                  (1a)
                sum_{k in K} c_k*w_k >= C                               (1b)
                xi_i >= 0, i in P                                       (1c)
                w_k in {0,1}, k in K                                    (1d)

The first sum in the objective is the penalty for the misclassified positive
examples, and the second part is the penalty for the misclassified negative
examples. Constraints (1a) ensure that each positive example either satisfies
at least one of the selected clause or the associated penalty variable is set
to 1. Constraints (1b) control the overall complexity of the generated DNF
formula. Clearly, the number of clauses is exponential, so the columns for w_k
are generated through a subproblem.

The reduced cost for a column $w_k$ is given by
sum_{i in mathcal{Z}} delta_i - sum_{i in mathcal{P}} mu_i delta_i + lambda
c_k where mu_i (>= 0) are the dual costs for constraints (1a) and lambda (>= 0)
is the dual cost for constraint (1c). The binary variable delta_i is set to 1
if the ith example satisfies the clause w_k.

Sets
P  set of all positive examples in the dataset.
Z  set of all negative examples in the dataset.
J  set of all the features.
S_i  the zero-valued features in sample i, S_i = {j in J: x_{ij} = 0}.

Parameters
D maximum complexity of the generated clause.

Decision Variables
z_j binary variable indicating if the feature j is present in the selected
    clause.
delta_i binary variable indicating if the input i is satisfied by the generated
        clause.

z_{CG} = min    lambda*(1 + sum_{j in J} z_j) -
                    - sum_{i in P} mu_i*delta_i + sum_{i in Z} delta_i  (obj)
         s.t.   delta_i + z_j <= 1, j in S_i, i in P                    (2a)
                delta_i >= 1 - sum_{j in S_i}z_j, i in Z                (2b)
                sum_{j in J} z_j <= D                                   (2c)
                z_j in {0,1}, j in J                                    (2d)
                delta_i >= 0, i in Z                                    (2e)

In the objective, the term 1 + sum_{j in J} z_j is equal to the generated
clause complexity, i.e., to 1 plus the clause size. This objective function
corresponds to the reduced cost of the clause generated. Negative objective
values identify clauses with a negative reduced cost, and the clause can be
added to the restricted master problem (RMP) in (1). The clause is constructed
by selecting the words for which the corresponding z variables are set to 1.
The constraints (2a) - (2b) relate the selected features with the delta
variables. Constraint (2c) controls the complexity of the clause being
generated.  Note that here the variables delta_i are binary but need not be
encoded as binary, as it is implied.
"""
import numpy as np
import itertools
from bitarray.util import int2ba
from ortools.linear_solver import pywraplp
from ortools.linear_solver import linear_solver_pb2
from sklearn.utils.validation import check_array, check_is_fitted
from ._col_gen_classifier import BaseMasterProblem
from ._col_gen_classifier import BaseSubproblem
from ._col_gen_classifier import ColGenClassifier


def get_params_from_string(params):
    """ Given the params in string form, returns the MPSolverParameters.
    Parameters
    ----------
    params : string,
        The solver parameters in string format.
    """
    # TODO: Implement this method.
    solver_params = pywraplp.MPSolverParameters()
    print(params)
    return solver_params


class BDRMasterProblem(BaseMasterProblem):
    """The master problem for boolean decision rule generation described in
    'Boolean decision rules via column generation' by Sanjeeb Dash et. al.
    2018. This extends the BaseMasterProblem for column generation classifier.

    Parameters
    ----------
    C : int, default=10,
        A parameter used for controlling the overall complexity of decision
        rule.
    p : float, default=1,
        A parameter used for balancing the penalty between false negatives and
        false positives.
        Higher value of p would result in more penalty for the false negatives.
    optimization_problem_type : string, default='glop',
        Describes the solver used for solving the master problem.

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`generate_mp`. The inputs should only
        contain values in {0,1}.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`generate_mp`. The labels should only
        contain values in {0,1}.
    solver_ : MPSolver from OR-Tools,
        The solver used for solving the master problem.
    clause_vars_ : list(int),
        Stores the indices of clause variables w_k generated so far.
    xi_vars_ : list(int),
        Stores the indices of positive penalty variables xi.
    clause_satisfaction_ : list(int)
        Stores the indices of constraints (1a).
    clause_complexity_: int
        Index of constraint (1b).
    clause_dict_ : dictionary, (int->list)
        Dictionary of clauses generated so far. The keys are the indices of the
        corresponding variables and the values are the lists containing the
        indices of features that make the clause.
    generated_ : boolean,
        True when the master problem model has been generated. False otherwise.
    """

    def __init__(self, C=10, p=1, optimization_problem_type='glop'):
        self.generated_ = False
        self.C_ = C
        self.p_ = p
        self.solver_ = pywraplp.Solver.CreateSolver(optimization_problem_type)

        # Vars
        self.clause_vars_ = None
        self.xi_vars_ = None

        # Constraints
        self.clause_satisfaction_ = None
        self.clause_complexity_ = None

        self.clause_dict_ = {}

    def generate_mp(self, X, y):
        """Generates the master problem model (RMP) and initializes the primal
        and dual solutions.
        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The input. The inputs should only contain values in {0,1}.
        y : ndarray, shape (n_samples,)
            The labels. The labels should only contain values in {0,1}.
        """
        if self.generated_:
            return

        self.X_ = X
        self.y__ = y
        infinity = self.solver_.infinity()
        # If not specified, Gurobi uses all the threads available. For solving
        # the LP, we only need one thread.
        self.solver_.SetNumThreads(1)

        n_positive_examples = len(list(filter(lambda x: (x > 0), y)))

        # Initial number of clasuses to start with.
        n_clauses = 1

        objective = self.solver_.Objective()
        self.xi_vars_ = [None]*n_positive_examples
        for i in range(n_positive_examples):
            # This need not be a boolean variable.
            # The upper bound is not needed.
            xi_var = self.solver_.IntVar(0.0, infinity, 'p_'+str(i))
            self.xi_vars_[i] = xi_var.index()
            objective.SetCoefficient(xi_var, self.p_)

        self.clause_vars_ = [None]*n_clauses
        for i in range(n_clauses):
            # This need not be a boolean variable.
            # The upper bound is not needed.
            clause_var = self.solver_.IntVar(0.0, infinity, 'c_'+str(i))
            self.clause_vars_[i] = clause_var.index()
            clause = self.generate_lexicographic_clause(i)
            self.clause_dict_[self.clause_vars_[i]] = clause
            obj_coeff = self.get_objective_coeff_mp(clause)
            objective.SetCoefficient(clause_var, obj_coeff)
        objective.SetMinimization()

        # Add constraints
        self.clause_satisfaction_ = [None]*n_positive_examples
        clause_satisfaction_cons = [None]*n_positive_examples
        clause_complexity_cons = self.solver_.Constraint(
            -infinity, self.C_, "clause_complexity")
        self.clause_complexity_ = clause_complexity_cons.index()
        for i in range(n_positive_examples):
            clause_satisfaction_cons[i] = self.solver_.Constraint(
                1, infinity, "clause_satisfaction_" + str(i))
            self.clause_satisfaction_[i] = clause_satisfaction_cons[i].index()
            xi_var = self.solver_.variable(self.xi_vars_[i])
            clause_satisfaction_cons[i].SetCoefficient(xi_var, 1)
        for i in range(n_clauses):
            clause_var = self.solver_.variable(self.clause_vars_[i])
            coeffs = self.get_clause_coeffs(
                self.clause_dict_[self.clause_vars_[i]])
            for j in range(n_positive_examples):
                clause_satisfaction_cons[j].SetCoefficient(
                    clause_var, coeffs[j])
            clause_complexity_cons.SetCoefficient(clause_var, coeffs[-1])

        self.generated_ = True

    def add_column(self, clause):
        """Adds the given column to the master problem model. Retuns True if
        the column is added to the master problem.

        Boolean decision rule generation is a special case where we take the
        clause as input instead of the column coefficients.

        Parameters
        ----------
        clause : list(int),
            The clause for which the column is to be generated.
        """
        assert self.generated_

        if len(clause) == 0:
            return False

        # Avoid adding duplicates.
        assert clause not in self.clause_dict_.values()

        n_clauses = len(self.clause_vars_)
        n_positive_examples = self.solver_.NumConstraints() - 1
        clause_var = self.solver_.IntVar(
            0.0, self.solver_.infinity(), 'c_'+str(n_clauses))
        self.clause_vars_.append(clause_var.index())
        self.clause_dict_[clause_var.index()] = clause
        obj_coeff = self.get_objective_coeff_mp(clause)
        self.solver_.Objective().SetCoefficient(clause_var, obj_coeff)

        # Constraint coeffs
        coeffs = self.get_clause_coeffs(clause)
        assert n_positive_examples == len(self.clause_satisfaction_)
        for j in range(n_positive_examples):
            clause_satisfaction_cons = self.solver_.constraint(
                self.clause_satisfaction_[j])
            clause_satisfaction_cons.SetCoefficient(clause_var, coeffs[j])
        clause_comp_constraint = self.solver_.constraint(
            self.clause_complexity_)
        clause_comp_constraint.SetCoefficient(clause_var, coeffs[-1])
        return True

    def solve_rmp(self, solver_params=''):
        """Solves the RMP with given solver params.
        Returns the dual costs of constraints (1b) and (1a) in a tuple.
        Parameters
        ----------
        solver_params : string, default='',
            The solver parameters for solving the RMP.
        """
        assert self.generated_

        # TODO(krunalp): Warm start LP
        # Solve lp
        result_status = self.solver_.Solve(
            get_params_from_string(solver_params))
        n_positive_examples = self.solver_.NumConstraints() - 1

        # TODO: Hide this under optional logging flag.
        print('Number of variables RMIP = %d' % self.solver_.NumVariables())
        print('Number of constraints RMIP = %d' %
              self.solver_.NumConstraints())
        if result_status == pywraplp.Solver.OPTIMAL:
            print('RMP Optimal objective value = %f' %
                  self.solver_.Objective().Value())

        # Dual costs
        clause_comp_cons = self.solver_.constraint(self.clause_complexity_)
        cc_dual = abs(clause_comp_cons.dual_value())
        cs_duals = []
        for i in range(n_positive_examples):
            clause_satisfaction_cons = self.solver_.constraint(
                self.clause_satisfaction_[i])
            cs_duals.append(clause_satisfaction_cons.dual_value())
        return (cc_dual, cs_duals)

    def solve_ip(self, solver_params=''):
        """Solves the integer RMP with given solver params.
        Returns True if the explanation is generated.
        Parameters
        ----------
        solver_params : string, default='',
            The solver parameters for solving the integer RMP.
        """
        assert self.generated_

        # We can use sat here since all the coefficients and variables are
        # integer. We can also use gurobi. But sat is 2-3 times faster.
        # TODO: Solver type should be a parameter.
        solver = pywraplp.Solver.CreateSolver("sat")

        # We have to load the model from LP solver.
        # This copy is not avoidable with OR-Tools since we are now switching
        # form solving LP to solving IP.
        model_proto = linear_solver_pb2.MPModelProto()
        self.solver_.ExportModelToProto(model_proto)
        solver.LoadModelFromProto(model_proto)

        num_xi_vars = len(self.xi_vars_)

        result_status = solver.Solve(get_params_from_string(solver_params))
        print('Problem solved in %f milliseconds' % solver.wall_time())
        self.explanation = []

        has_solution = (
            result_status == pywraplp.Solver.OPTIMAL or
            result_status == pywraplp.Solver.FEASIBLE)
        assert has_solution

        # TODO: Store solution instead of printing it.
        print("Integer RMP Objective = ", solver.Objective().Value())
        all_vars = solver.variables()
        assert len(all_vars) == num_xi_vars + len(self.clause_vars_)
        for i in range(num_xi_vars, len(all_vars)):
            if all_vars[i].solution_value() > 0:
                orig_var = self.solver_.variable(
                    self.clause_vars_[i-num_xi_vars])
                self.explanation.append(self.clause_dict_[orig_var.index()])
        return True

    @staticmethod
    def satisfies_clause(entry, clause):
        """ Given the entry and clause, returns 1 if the entry satisfies the
        clause and 0 oterwise.
        Parameters
        ----------
        entry : ndarray, shape(n_features)
            The input example. The array should have values in {0,1}.
        clause : list(int),
            The list containing the indices of features present in the clause.
        """
        for index in clause:
            if entry[index] == 0:
                return 0
        return 1

    def get_objective_coeff_mp(self, clause):
        """ Returns the number of zero examples satisfying the clause.
        Parameters
        ----------
        clause : list(int),
            The list containing the indices of features present in the clause.
        """
        num_entries_satisfying_clause = 0
        for i in range(len(self.X_)):
            entry = self.X_[i]
            target = self.y__[i]
            # Only consider zero examples.
            if target == 1:
                continue
            num_entries_satisfying_clause += self.satisfies_clause(
                entry, clause)
        return num_entries_satisfying_clause

    def get_clause_coeffs(self, clause):
        """ Given the clause, returns the coefficients for corresponding
        variable in RMP.
        Parameters
        ----------
        clause : list(int),
            The list containing the indices of features present in the clause.
        """
        # no of rows = No of positive examples + 1.
        coeffs = []
        for i in range(len(self.X_)):
            entry = self.X_[i]
            target = self.y__[i]
            # Only consider positive examples.
            if target == 0:
                continue
            coeffs.append(self.satisfies_clause(entry, clause))
        # last constraint
        coeffs.append(len(clause) + 1)
        return coeffs

    @staticmethod
    def generate_lexicographic_clause(index):
        """ Generates the 'index'th clause as per lexicogrphical index.
        Parameters
        ----------
        index : int.
        """
        ind = int2ba(index)
        ind = ind[::-1]
        clause = [i for i in range(len(ind)) if ind[i]]
        return clause


class BDRSubProblem(BaseSubproblem):
    """The  subproblem for boolean decision rule generation described in
    'Boolean decision rules via column generation' by Sanjeeb Dash et. al.
    2018. This extends the BaseSubproblem for column generation classifier.

    Parameters
    ----------
    D : int,
        A parameter used for controlling the complexity of the clause being
        generated.
    optimization_problem_type : string, default='cbc'
        Describes the solver used for solving the subproblem (2).

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`generate_columns`. The inputs should
        only contain values in {0,1}.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`generate_columns`. The labels should
        only contain values in {0,1}.
    solver_ : MPSolver from OR-Tools,
        The solver used for solving the subproblem.
    delta_vars_ : list(int),
        Stores the indices of delta variables in (2).
    z_vars_ : list(int),
        Stores the indices of z variables in (2).
    generated_ : boolean,
        True when the master problem model has been generated. False otherwise.
    """

    def __init__(self, D, optimization_problem_type='cbc'):
        self.D_ = D
        self.solver_ = pywraplp.Solver.CreateSolver(optimization_problem_type)
        self.generated_ = False

        # Vars
        self.delta_vars_ = None
        self.z_vars_ = None

    def create_submip(self, cc_dual, cs_duals):
        """Creates the model for the subproblem. This should be called only
        once for a given problem.
        Parameters
        ----------
        cc_dual : float,
            dual cost of clause complexity constraint (1b).
        cs_duals : list(float),
            dual costs of clause satisfaction constraints (1a).
        """
        assert not self.generated_, "SP is already created."
        infinity = self.solver_.infinity()

        n_words = self.X_.shape[1]
        n_samples = self.X_.shape[0]

        self.z_vars_ = [None]*n_words
        objective = self.solver_.Objective()
        for i in range(n_words):
            z_var = self.solver_.BoolVar('z_'+str(i))
            self.z_vars_[i] = z_var.index()
            objective.SetCoefficient(z_var, cc_dual)

        self.delta_vars_ = [None]*n_samples
        n_positive_examples = 0
        for i in range(n_samples):
            # This need not be a boolean variable.
            delta_var = self.solver_.BoolVar('d_'+str(i))
            self.delta_vars_[i] = delta_var.index()
            obj_coeff = 1
            target = self.y_[i]
            if target == 1:
                obj_coeff *= -cs_duals[n_positive_examples]
                n_positive_examples += 1
            objective.SetCoefficient(delta_var, obj_coeff)

        objective.SetOffset(cc_dual)
        objective.SetMinimization()

        # Constraints
        for i in range(n_samples):
            delta_var = self.solver_.variable(self.delta_vars_[i])
            if self.y_[i] == 1:
                for j in range(n_words):
                    if self.X_[i, j] == 0:
                        cons = self.solver_.Constraint(
                            0, 1, "pos_" + str(i) + "_" + str(j))
                        z_var = self.solver_.variable(self.z_vars_[j])
                        cons.SetCoefficient(delta_var, 1)
                        cons.SetCoefficient(z_var, 1)
            else:
                cons = self.solver_.Constraint(1, infinity, "zero_" + str(i))
                cons.SetCoefficient(delta_var, 1)
                for j in range(n_words):
                    if self.X_[i, j] == 0:
                        z_var = self.solver_.variable(self.z_vars_[j])
                        cons.SetCoefficient(z_var, 1)

        cc_const = self.solver_.Constraint(0, self.D_, "clause_complexity")
        for j in range(n_words):
            z_var = self.solver_.variable(self.z_vars_[j])
            cc_const.SetCoefficient(z_var, 1)
        self.generated_ = True

    def update_objective(self, cc_dual, cs_duals):
        """Updates the objective of the generated subproblem. This can be
        called only after the create_submip method has been called.
        Parameters
        ----------
        cc_dual : float,
            dual cost of clause complexity constraint (1b).
        cs_duals : list(float),
            dual costs of clause satisfaction constraints (1a).
        """
        assert self.generated_, "Subproblem not generated."
        n_words = self.X_.shape[1]
        n_samples = self.X_.shape[0]
        objective = self.solver_.Objective()
        for i in range(n_words):
            z_var = self.solver_.variable(self.z_vars_[i])
            objective.SetCoefficient(z_var, cc_dual)

        n_positive_examples = 0
        for i in range(n_samples):
            obj_coeff = 1
            target = self.y_[i]
            if target == 1:
                obj_coeff *= -cs_duals[n_positive_examples]
                n_positive_examples += 1
            delta_var = self.solver_.variable(self.delta_vars_[i])
            objective.SetCoefficient(delta_var, obj_coeff)

        objective.SetOffset(cc_dual)
        objective.SetMinimization()

    def generate_columns(self, X, y, dual_costs, params=""):
        """Generates the new columns to be added to the RMP.
        In this case instead of directly generating the coefficients, this
        method returns the list of generated clauses. The Master problem can
        find the coefficients from it.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The input. The inputs should only contain values in {0,1}.
        y : ndarray, shape (n_samples,)
            The labels. The labels should only contain values in {0,1}.
        dual_costs : tuple(float), size=2
            Dual costs of constraints (1b) and (1a) in a tuple.
        params : string, default=""
            Solver parameters.
        """
        self.X_ = X
        self.y_ = y
        cc_dual = dual_costs[0]
        cs_duals = dual_costs[1]
        # We assume the positive dual cost in this model.
        assert cc_dual >= 0, "Negative clause complexity dual"
        for cs_dual in cs_duals:
            assert cs_dual >= 0, "Negative clause satisfaction dual"

        if self.generated_:
            self.update_objective(cc_dual, cs_duals)
        else:
            self.create_submip(cc_dual, cs_duals)

        # Solve sub problem
        result_status = self.solver_.Solve(get_params_from_string(params))

        # Empty column is always feasible.
        has_solution = (result_status == pywraplp.Solver.OPTIMAL or
                        result_status == pywraplp.Solver.FEASIBLE)
        assert has_solution

        # Try to generate the clause if negative objective
        # TODO: This threshold should be a parameter.
        if self.solver_.Objective().Value() >= -1e-6:
            return []

        clause = []
        for i in range(len(self.z_vars_)):
            z_var = self.solver_.variable(self.z_vars_[i])
            if z_var.solution_value() > 0:
                clause.append(i)

        return [clause]


class BooleanDecisionRuleClassifier(ColGenClassifier):
    """Binary classifier using boolean decision rule generation method.

    Parameters
    ----------
    max_iterations : int, default=-1
        Maximum column generation iterations. Negative values removes the
        iteration limit and the problem is solved till optimality.
    C : int,default=10,
        A parameter used for controlling the overall complexity of decision
        rule.
    p : float,default=1
        A parameter used for balancing the penalty between false negatives and
        false positives. Higher value of p would result in more penalty for the
        false negatives.
    rmp_solver_params: string, default = "",
        Solver parameters for solving restricted master problem (rmp).
    rmp_solver: string, default = "glop",
        Solver used for solving restricted master problem (rmp).
    master_ip_solver_params: string, default = "",
        Solver parameters for solving the integer master problem.
    subproblem_solver: string, default = "cbc",
        Solver used for solving subproblem.
    subproblem_params: list of strings, default = [""],
        Parameters for solving the subproblem.

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`fit`.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """

    def __init__(self, max_iterations=-1,
                 C=10,
                 p=1,
                 rmp_solver_params="",
                 rmp_solver="glop",
                 master_ip_solver_params="",
                 subproblem_solver="cbc",
                 subproblem_params=[""]):
        super(BooleanDecisionRuleClassifier, self).__init__(
            max_iterations,
            master_problem=BDRMasterProblem(
                C, p, rmp_solver),
            subproblems=[BDRSubProblem(
                C, subproblem_solver)],
            rmp_is_ip=True,
            rmp_solver_params=rmp_solver_params,
            master_ip_solver_params=master_ip_solver_params,
            subproblem_params=subproblem_params)
        # Not used after this but we have to store them for sklearn
        # compatibility.
        self.C = C
        self.p = p
        self.rmp_solver = rmp_solver
        self.subproblem_solver = subproblem_solver

    def _more_tags(self):
        return {'X_types': ['categorical'],
                'non_deterministic': True,
                'binary_only': True}

    def predict(self, X):
        """Predicts the class based on the solution of master problem. This
        method is abstract.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples. The inputs should only contain values in {0,1}.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample. The labels only contain values in {0,1}.
        """
        X = check_array(X, accept_sparse=True)
        check_is_fitted(self, 'is_fitted_')
        explanation = self.master_problem.explanation
        print("Explanation=", explanation)
        # Check if the input satisfies any of the clause in explanation.
        y_predict = np.zeros(X.shape[0], dtype=int)
        for i in range(X.shape[0]):
            for clause in explanation:
                if self.master_problem.satisfies_clause(X[i], clause):
                    y_predict[i] = 1
                    break
        # Translate y back to original form.
        return self.label_encoder_.inverse_transform(y_predict)


class BDRHeuristic(BaseSubproblem):
    """Heuristically generates the new columns to be added to the RMP.
       Iterates through all size K clauses and returns first one with the
       negative reduced cost.
       This extends the BaseSubproblem for column generation classifier.

    Parameters
    ----------
    K : int,
        maximum length of generated clauses.

    Attributes
    ----------

    """

    def __init__(self, K):
        self.K_ = K

    def generate_columns(self, X, y, dual_costs, params=""):
        """Heuristically generates the new columns to be added to the RMP.
        Iterates through all size K clauses and returns all that have the
        negative reduced cost.

        In this case instead of directly generating the coefficients, this
        method returns the list of generated clauses. The Master problem can
        find the coefficients from it.

        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The input. The inputs should only contain values in {0,1}.
        y : ndarray, shape (n_samples,)
            The labels. The labels should only contain values in {0,1}.
        dual_costs : tuple(float), size=2
            Dual costs of constraints (1b) and (1a) in a tuple.
        params : string, default=""
            unused parameters.
        """
        cc_dual = dual_costs[0]
        cs_duals = dual_costs[1]
        # We assume the positive dual cost in this model.
        assert cc_dual >= 0, "Negative clause complexity dual"
        for cs_dual in cs_duals:
            assert cs_dual >= 0, "Negative clause satisfaction dual"

        n_words = X.shape[1]
        n_samples = X.shape[0]
        # Iterate through all size K clause.
        all_words = range(n_words)
        neg_red_cost_clauses = []
        for clause_size in range(1, self.K_+1):
            clauses = itertools.combinations(all_words, clause_size)
            for clause in clauses:
                reduced_cost = cc_dual * (1 + len(clause))
                pos_ind = -1
                for j in range(n_samples):
                    entry = X[j]
                    target = y[j]
                    if target == 1:
                        pos_ind += 1
                    satisfied = BDRMasterProblem.satisfies_clause(
                        entry, clause)
                    if satisfied == 0:
                        continue

                    if target == 1:
                        reduced_cost -= cs_duals[pos_ind]
                    else:
                        reduced_cost += 1
                if reduced_cost < -1e-6:
                    neg_red_cost_clauses.append(list(clause))
        return neg_red_cost_clauses


class BooleanDecisionRuleClassifierWithHeuristic(
        BooleanDecisionRuleClassifier):
    """Binary classifier using boolean decision rule generation method with
    heuristic for column generation.

    Parameters
    ----------
    max_iterations : int, default=-1
        Maximum column generation iterations. Negative values removes the
        iteration limit and the problem is solved till optimality.
    C : int,default=10,
        A parameter used for controlling the overall complexity of decision
        rule.
    p : float,default=1
        A parameter used for balancing the penalty between false negatives and
        false positives. Higher value of p would result in more penalty for the
        false negatives.
    rmp_solver_params: string, default = "",
        Solver parameters for solving restricted master problem (rmp).
    rmp_solver: string, default = "glop",
        Solver used for solving restricted master problem (rmp).
    master_ip_solver_params: string, default = "",
        Solver parameters for solving the integer master problem.
    subproblem_solver: string, default = "cbc",
        Solver used for solving subproblem.
    subproblem_params: list of strings, default = ["",""],
        Parameters for solving the subproblem. First string is for the
        heuristic method and the second is for the integer program.

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`fit`.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    """

    def __init__(self, max_iterations=-1,
                 C=10,
                 p=1,
                 rmp_solver_params="",
                 rmp_solver="glop",
                 master_ip_solver_params="",
                 subproblem_solver="cbc",
                 subproblem_params=["", ""]):
        super(BooleanDecisionRuleClassifier, self).__init__(
            max_iterations,
            master_problem=BDRMasterProblem(
                C, p, rmp_solver),
            subproblems=[BDRHeuristic(K=1),
                         BDRSubProblem(
                C, subproblem_solver)],
            rmp_is_ip=True,
            rmp_solver_params=rmp_solver_params,
            master_ip_solver_params=master_ip_solver_params,
            subproblem_params=subproblem_params)
        # Not used after this but we have to store them for sklearn
        # compatibility.
        self.C = C
        self.p = p
        self.rmp_solver = rmp_solver
        self.subproblem_solver = subproblem_solver



"""
TODO: Documentation of the file.
"""
import numpy as np
import math
import itertools
from bitarray.util import int2ba
from ortools.linear_solver import pywraplp
from ortools.linear_solver import linear_solver_pb2
from sklearn.utils.validation import check_array, check_is_fitted
from ._col_gen_classifier import BaseMasterProblem
from ._col_gen_classifier import BaseSubproblem
from ._col_gen_classifier import ColGenClassifier


class Path:
    """TODO: Documentation.
    """

    def __init__(self) -> None:
        self.leaf_id = -1
        self.node_ids = []
        self.splits = []
        self.cost = 0
        self.id = -1
        self.target = -1

    def set_leaf(self, leaf_id, depth):
        """TODO: Documentation.
        For a full binary tree, adds the nodes corresponding to the given
        leaf_id."""
        self.leaf_id = leaf_id
        self.node_ids = []
        count = 2**(depth) - 1 + leaf_id
        while(count > 0):
            father = math.floor((count-1) / 2)
            self.node_ids.append(father)
            count = father


class Node:
    """TODO: Documentation.
    """

    def __init__(self) -> None:
        self.id = -1
        self.candidate_splits = []


class Leaf:
    """TODO: Documentation.
    """

    def __init__(self) -> None:
        self.id = id
        self.left_nodes = []
        self.right_nodes = []

    def create_leaf(self, id, depth) -> None:
        """TODO: Documentation.
        """
        self.id = id
        self.left_nodes = []
        self.right_nodes = []
        count = 2**(depth) - 1 + id
        while(count > 0):
            father = math.floor((count-1) / 2)
            if count % 2 == 0:
                self.right_nodes.append(father)
            else:
                self.left_nodes.append(father)
            count = father


class Split:
    """TODO: Documentation.
    """

    def __init__(self) -> None:
        self.id = -1
        self.feature = -1
        self.threshold = -1


def row_satisfies_path(X, leaf, splits, row, path):
    """TODO: Documentation.
    """
    split_ids = path.splits
    node_ids = path.node_ids
    for i in range(len(node_ids)):
        split = splits[split_ids[i]]
        feature = split.feature
        threshold = split.threshold
        if node_ids[i] in leaf.left_nodes:
            if X[row, feature] > threshold:
                return False
        else:  # Right node in the path
            if X[row, feature] <= threshold:
                return False
    return True


def get_params_from_string(params):
    """ Given the params in string form, returns the MPSolverParameters.
    Parameters
    ----------
    params : string,
        The solver parameters in string format.
    """
    # TODO: Implement this method.
    solver_params = pywraplp.MPSolverParameters()
    print(params)
    return solver_params


class DTreeMasterProblem(BaseMasterProblem):
    """TODO: Documentation.
    """

    def __init__(self, initial_paths, leaves, nodes, splits):
        super().__init__()
        self.solver_ = pywraplp.Solver.CreateSolver('glop')
        self.generated_ = False
        self.paths_ = initial_paths
        self.leaves_ = leaves
        self.nodes_ = nodes
        self.splits_ = splits

    def generate_mp(self, X, y):
        """TODO: Documentation.
        """
        if self.generated_:
            return

        self.X_ = X
        self.y__ = y
        infinity = self.solver_.infinity()
        self.solver_.SetNumThreads(1)

        objective = self.solver_.Objective()
        self.path_vars_ = [None]*len(self.paths_)
        for i in range(len(self.paths_)):
            xp_var = self.solver_.IntVar(0.0, infinity, 'p_'+str(i))
            self.path_vars_[i] = xp_var.index()
            self.paths_[i].id = xp_var.index()
            objective.SetCoefficient(xp_var, self.paths_[i].cost)
        objective.SetMaximization()

        # leaf constraints
        self.leaf_cons_ = [None]*len(self.leaves_)
        for leaf in self.leaves_:
            self.leaf_cons_[leaf.id] = self.solver_.Constraint(
                1, 1, "leaf_"+str(leaf.id))
            for path in self.paths_:
                if path.leaf_id == leaf.id:
                    xp_var = self.solver_.variable(path.id)
                    self.leaf_cons_[leaf.id].SetCoefficient(xp_var, 1)

        # row constraints
        n_rows = self.X_.shape[0]
        self.row_cons_ = [None]*n_rows
        for r in range(n_rows):
            self.row_cons_[r] = self.solver_.Constraint(1, 1, "row_"+str(r))
            for path in self.paths_:
                if row_satisfies_path(X, self.leaves_[path.leaf_id],
                                      self.splits_, r, path):
                    xp_var = self.solver_.variable(path.id)
                    self.row_cons_[r].SetCoefficient(xp_var, 1)

        # consistency constraints
        self.ns_constraints_ = {}
        for leaf in self.leaves_:
            self.ns_constraints_[leaf.id] = {}
            nodes = leaf.left_nodes + leaf.right_nodes
            for node_id in nodes:
                node = self.nodes_[node_id]
                self.ns_constraints_[leaf.id][node.id] = {}
                for split in node.candidate_splits:

                    ns_constraint = self.solver_.Constraint(
                        0, 0, "ns_"+str(node.id)+"_"+str(split)+"_"
                        + str(leaf.id))

                    dummy_var = self.solver_.BoolVar("r_ns_" +
                                                     str(node.id) +
                                                     "_"+str(split) +
                                                     "_"+str(leaf.id))
                    ns_constraint.SetCoefficient(dummy_var, -1)

                    for path in self.paths_:
                        for i in range(len(path.node_ids)):
                            if path.leaf_id == leaf.id\
                                and path.node_ids[i] == node.id \
                                    and path.splits[i] == split:
                                xp_var = self.solver_.variable(path.id)
                                ns_constraint.SetCoefficient(xp_var, 1)

                    self.ns_constraints_[
                        leaf.id][node.id][split] = ns_constraint
        # print(self.solver_.ExportModelAsLpFormat(False))
        self.generated_ = True

    def add_column(self, path):
        """TODO: Documentation.
        """
        objective = self.solver_.Objective()
        infinity = self.solver_.infinity()
        xp_var = self.solver_.IntVar(
            0.0, infinity, 'p_'+str(len(self.path_vars_)))
        self.path_vars_.append(xp_var.index())
        path.id = xp_var.index()
        objective.SetCoefficient(xp_var, path.cost)
        self.paths_.append(path)

        # leaf constraints
        for leaf in self.leaves_:
            if path.leaf_id == leaf.id:
                self.leaf_cons_[leaf.id].SetCoefficient(xp_var, 1)

        # row constraints
        n_rows = self.X_.shape[0]
        for r in range(n_rows):
            if row_satisfies_path(self.X_, self.leaves_[path.leaf_id],
                                  self.splits_, r, path):
                self.row_cons_[r].SetCoefficient(xp_var, 1)

        # consistency constraints
        for leaf in self.leaves_:
            nodes = leaf.left_nodes + leaf.right_nodes
            for node_id in nodes:
                node = self.nodes_[node_id]
                for split in node.candidate_splits:

                    ns_constraint = self.ns_constraints_[
                        leaf.id][node.id][split]

                    for i in range(len(path.node_ids)):
                        if path.leaf_id == leaf.id\
                            and path.node_ids[i] == node.id \
                                and path.splits[i] == split:
                            ns_constraint.SetCoefficient(xp_var, 1)
        return True

    def solve_rmp(self, solver_params=''):
        """TODO: Documentation.
        """
        assert self.generated_

        result_status = self.solver_.Solve(
            get_params_from_string(solver_params))
        n_paths = len(self.paths_)

        # TODO: Hide this under optional logging flag.
        print('Number of variables RMIP = %d' % self.solver_.NumVariables())
        print('Number of constraints RMIP = %d' %
              self.solver_.NumConstraints())
        if result_status == pywraplp.Solver.OPTIMAL:
            print('RMP Optimal objective value = %f' %
                  self.solver_.Objective().Value())

            for var in self.solver_.variables():
                print(var.name(), var.solution_value())

            leaf_duals = []
            for lid in range(len(self.leaves_)):
                leaf_duals.append(self.leaf_cons_[lid].dual_value())

            row_duals = []
            n_rows = self.X_.shape[0]
            for r in range(n_rows):
                row_duals.append(self.row_cons_[r].dual_value())

            ns_duals = {}
            for leaf in self.leaves_:
                ns_duals[leaf.id] = {}
                nodes = leaf.left_nodes + leaf.right_nodes
                for node_id in nodes:
                    node = self.nodes_[node_id]
                    ns_duals[leaf.id][node.id] = {}
                    for split in node.candidate_splits:
                        ns_duals[leaf.id][node.id][split] = \
                            self.ns_constraints_[
                            leaf.id][node.id][split].dual_value()
        else:
            print(result_status)
            print(self.solver_.ExportModelAsLpFormat(False))

        return (leaf_duals, row_duals, ns_duals)

    def solve_ip(self, solver_params=''):
        """Solves the integer RMP with given solver params.
        Returns True if the tree is generated.
        Parameters
        ----------
        solver_params : string, default='',
            The solver parameters for solving the integer RMP.
        """
        assert self.generated_

        # We can use sat here since all the coefficients and variables are
        # integer.
        # TODO: Solver type should be a parameter.
        solver = pywraplp.Solver.CreateSolver("sat")

        # We have to load the model from LP solver.
        # This copy is not avoidable with OR-Tools since we are now switching
        # form solving LP to solving IP.
        model_proto = linear_solver_pb2.MPModelProto()
        self.solver_.ExportModelToProto(model_proto)
        solver.LoadModelFromProto(model_proto)

        num_path_vars = len(self.path_vars_)

        result_status = solver.Solve(get_params_from_string(solver_params))
        print('Problem solved in %f milliseconds' % solver.wall_time())
        self.selected_paths = []

        has_solution = (
            result_status == pywraplp.Solver.OPTIMAL or
            result_status == pywraplp.Solver.FEASIBLE)
        assert has_solution

        # TODO: Store solution instead of printing it.
        print("Integer RMP Objective = ", solver.Objective().Value())
        all_vars = solver.variables()

        for path in self.paths_:
            path_var = solver.variable(path.id)
            if path_var.solution_value() > 0:
                self.selected_paths.append(path)
        return True


class DTreeSubProblem(BaseSubproblem):
    """TODO: Documentation."""

    def __init__(self, leaf, nodes, splits, targets,
                 depth, optimization_problem_type='cbc') -> None:
        super().__init__()
        self.leaf_id_ = leaf.id
        self.leaf_ = leaf
        self.nodes_ = nodes
        self.splits_ = splits
        self.targets_ = targets
        self.tree_depth_ = depth
        self.solver_ = pywraplp.Solver.CreateSolver(optimization_problem_type)
        self.generated_ = False
        self.z_vars_ = None

    def create_submip(self, leaf_dual, row_duals, ns_duals):
        """TODO: Documentation.
        """
        assert not self.generated_, "SP is already created."
        infinity = self.solver_.infinity()

        n_rows = self.X_.shape[0]
        self.z_vars_ = [None]*n_rows
        self.y_vars_ = [None]*n_rows
        objective = self.solver_.Objective()
        for i in range(n_rows):
            z_var = self.solver_.BoolVar('z_'+str(i))
            objective.SetCoefficient(z_var, 1)
            self.z_vars_[i] = z_var.index()

            y_var = self.solver_.BoolVar('y_'+str(i))
            row_dual = row_duals[i]
            objective.SetCoefficient(y_var, -row_dual)
            self.y_vars_[i] = y_var.index()

        self.u_vars_ = {}
        self.leaf_
        nodes = self.leaf_.left_nodes + self.leaf_.right_nodes
        all_splits = []
        for node_id in nodes:
            node = self.nodes_[node_id]
            self.u_vars_[node.id] = {}
            for split in node.candidate_splits:
                all_splits.append(split)
                u_var = self.solver_.BoolVar(
                    'u_'+str(node.id)+'_'+str(split))
                ns_dual = ns_duals[self.leaf_id_][node.id][split]
                objective.SetCoefficient(u_var, -ns_dual)
                self.u_vars_[node.id][split] = u_var.index()

        objective.SetOffset(-leaf_dual)
        objective.SetMaximization()

        # Constraints
        # For each node, exactly one feature is selected
        for node_id in nodes:
            node = self.nodes_[node_id]
            one_feature = self.solver_.Constraint(
                1, 1, "one_feature_" + str(node.id))
            for split in node.candidate_splits:
                u_var = self.solver_.variable(self.u_vars_[node.id][split])
                one_feature.SetCoefficient(u_var, 1)

        # Each feature is selected at max once in a path
        all_splits = list(dict.fromkeys(all_splits))
        for split in all_splits:
            unique_split = self.solver_.Constraint(
                0, 1, "unique_split_" + str(split))
            for node in self.nodes_:
                if node.id not in nodes:
                    continue
                if split not in node.candidate_splits:
                    continue
                u_var = self.solver_.variable(self.u_vars_[node.id][split])
                unique_split.SetCoefficient(u_var, 1)

        # Row follows correct path (upper and lower bound on y var)
        for i in range(n_rows):
            y_lb_cons = self.solver_.Constraint(
                -infinity, self.tree_depth_-1, "row_" + str(i))
            y_var = self.solver_.variable(self.y_vars_[i])
            y_lb_cons.SetCoefficient(y_var, -1)
            for node in self.nodes_:
                if node.id not in nodes:
                    continue
                rn_cons = self.solver_.Constraint(
                    -infinity, 0, "rn_" + str(i) + '_' + str(node.id))

                rn_cons.SetCoefficient(y_var, 1)
                if node.id in self.leaf_.left_nodes:
                    for split in node.candidate_splits:
                        if self.row_satisfies_split(i, self.splits_[split]):
                            u_var = self.solver_.variable(
                                self.u_vars_[node.id][split])
                            rn_cons.SetCoefficient(u_var, -1)
                            y_lb_cons.SetCoefficient(u_var, 1)
                elif node.id in self.leaf_.right_nodes:
                    for split in node.candidate_splits:
                        if not self.row_satisfies_split(i,
                                                        self.splits_[split]):
                            u_var = self.solver_.variable(
                                self.u_vars_[node.id][split])
                            rn_cons.SetCoefficient(u_var, -1)
                            y_lb_cons.SetCoefficient(u_var, 1)

        # only one class selected
        self.w_vars_ = [None]*len(self.targets_)
        single_target_cons = self.solver_.Constraint(0, 1, 'single_target')
        for i in range(len(self.targets_)):
            target = self.targets_[i]
            w_var = self.solver_.BoolVar('w_' + str(target))
            self.w_vars_[i] = w_var.index()
            single_target_cons.SetCoefficient(w_var, 1)
        # relation between w,y,z vars
        for i in range(n_rows):
            z_var = self.solver_.variable(self.z_vars_[i])
            y_var = self.solver_.variable(self.y_vars_[i])
            yz_cons = self.solver_.Constraint(-infinity, 0, "yz_" + str(i))
            yz_cons.SetCoefficient(z_var, 1)
            yz_cons.SetCoefficient(y_var, -1)
            correct_class = self.y_[i]
            w_var = self.solver_.variable(self.w_vars_[correct_class])
            wz_cons = self.solver_.Constraint(-infinity, 0, "wz_" + str(i))
            wz_cons.SetCoefficient(z_var, 1)
            wz_cons.SetCoefficient(w_var, -1)

        self.generated_ = True

    def update_objective(self, leaf_dual, row_duals, ns_duals):
        """TODO: Documentation.
        """
        objective = self.solver_.Objective()
        n_rows = self.X_.shape[0]
        for i in range(n_rows):
            z_var = self.solver_.variable(self.z_vars_[i])
            objective.SetCoefficient(z_var, 1)

            y_var = self.solver_.variable(self.y_vars_[i])
            row_dual = row_duals[i]
            objective.SetCoefficient(y_var, -row_dual)

        nodes = self.leaf_.left_nodes + self.leaf_.right_nodes
        for node_id in nodes:
            node = self.nodes_[node_id]
            for split in node.candidate_splits:
                u_var = self.solver_.variable(self.u_vars_[node.id][split])
                ns_dual = ns_duals[self.leaf_id_][node.id][split]
                objective.SetCoefficient(u_var, -ns_dual)

        objective.SetOffset(-leaf_dual)
        objective.SetMaximization()
        return

    def generate_columns(self, X, y, dual_costs, params=""):
        """TODO: Documentation.
        """
        self.X_ = X
        self.y_ = y
        leaf_dual = dual_costs[0][self.leaf_id_]
        row_duals = dual_costs[1]
        ns_duals = dual_costs[2]

        self.X_ = X
        self.y__ = y
        self.solver_.SetNumThreads(1)

        if self.generated_:
            self.update_objective(leaf_dual, row_duals, ns_duals)
        else:
            self.create_submip(leaf_dual, row_duals, ns_duals)

        # Solve sub problem
        result_status = self.solver_.Solve(get_params_from_string(params))

        # Empty column is always feasible.
        has_solution = (result_status == pywraplp.Solver.OPTIMAL or
                        result_status == pywraplp.Solver.FEASIBLE)
        assert has_solution

        # print(self.solver_.ExportModelAsLpFormat(False))

        # TODO: This threshold should be a parameter.
        print(self.solver_.Objective().Value())
        if self.solver_.Objective().Value() <= 1e-6:
            return []

        nodes = self.leaf_.left_nodes + self.leaf_.right_nodes
        path = Path()
        path.leaf_id = self.leaf_id_
        path.node_ids = []
        path.splits = []
        path.cost = 0
        path.id = -1
        for node in self.nodes_:
            if node.id not in nodes:
                continue
            path.node_ids.append(node.id)
            for split in node.candidate_splits:
                u_var = self.solver_.variable(self.u_vars_[node.id][split])
                if u_var.solution_value() > 0.5:
                    path.splits.append(split)
                    break
        for target in self.targets_:
            w_var = self.solver_.variable(self.w_vars_[target])
            if w_var.solution_value() > 0.5:
                path.target = target
                break
        n_rows = self.X_.shape[0]
        for i in range(n_rows):
            z_var = self.solver_.variable(self.z_vars_[i])
            path.cost += z_var.solution_value()
            if z_var.solution_value() > 0.5:
                print(self.z_vars_[i], z_var)
        return [path]

    def row_satisfies_split(self, row, split):
        """TODO: Documentation.
        """
        feature = split.feature
        threshold = split.threshold
        if self.X_[row, feature] <= threshold:
            return True
        return False


class DTreeClassifier(ColGenClassifier):
    """TODO: Documentation.
    """

    def __init__(self, initial_paths=[], leaves=[], nodes=[], splits=[],
                 tree_depth=1,
                 targets=[], max_iterations=-1, rmp_is_ip=True,
                 rmp_solver_params="",
                 master_ip_solver_params="", subproblem_param_str=""):

        self.leaves_ = leaves
        self.initial_paths_ = initial_paths
        self.nodes_ = nodes
        self.splits_ = splits
        self.targets_ = targets
        self.tree_depth_ = tree_depth
        split_ids = []
        for split in splits:
            split_ids.append(split.id)
        # Each node must be in sequence.
        node_ids = []
        node_id = 0
        for node in nodes:
            assert node.id == node_id
            node_ids.append(node.id)
            node_id += 1

        # Each leaf must be in sequence.
        leaf_ids = []
        leaf_id = 0
        for leaf in leaves:
            assert leaf.id == leaf_id
            leaf_ids.append(leaf.id)
            leaf_id += 1
        # Each path can only contain nodes and leaves provided.
        for path in self.initial_paths_:
            assert path.leaf_id in leaf_ids
            for node_id in path.node_ids:
                assert node_id in node_ids
            for split_id in path.splits:
                assert split_id in split_ids
            assert path.target in targets
            assert len(path.node_ids) == self.tree_depth_
            assert len(path.splits) == self.tree_depth_

        self.master_problem = DTreeMasterProblem(
            self.initial_paths_, leaves, nodes, splits)
        self.subproblems = []
        subproblem_params = []
        for leaf in leaves:
            subproblem = DTreeSubProblem(
                leaf, nodes, splits, targets, self.tree_depth_, 'cbc')
            self.subproblems.append(subproblem)
            subproblem_params.append(subproblem_param_str)
        super().__init__(max_iterations, self.master_problem, self.subproblems,
                         rmp_is_ip, rmp_solver_params, master_ip_solver_params,
                         subproblem_params)

    def _more_tags(self):
        """TODO: Documentation.
        """
        return {'X_types': ['categorical'],
                'non_deterministic': True,
                'binary_only': True}

    def predict(self, X):
        """TODO: Documentation.
        """
        selected_paths = self.master_problem.selected_paths
        # Check for each row, which path it satisfies. There should be exactly
        # one.
        num_rows = X.shape[0]
        y_predict = np.zeros(X.shape[0], dtype=int)
        for row in range(num_rows):
            for path in selected_paths:
                if row_satisfies_path(X, self.leaves_[path.leaf_id],
                                      self.splits_, row, path):
                    y_predict[row] = path.target
                    break
        return self.label_encoder_.inverse_transform(y_predict)



import pytest
import numpy as np

from numpy.testing import assert_array_equal
from numpy.testing import assert_equal
from numpy.testing import assert_raises

from col_gen_estimator import BooleanDecisionRuleClassifier
from col_gen_estimator import BooleanDecisionRuleClassifierWithHeuristic
from col_gen_estimator import BDRMasterProblem
from col_gen_estimator import BDRSubProblem
from col_gen_estimator import BDRHeuristic


@pytest.fixture
def data():
    X = np.array([[1, 1, 1, 0],
                  [1, 1, 0, 1],
                  [0, 1, 1, 1],
                  [1, 0, 1, 1]], np.int8)
    y = np.array([1, 1, 0, 0], np.int8)
    return (X, y)


@pytest.fixture
def data_one_class():
    X = np.array([[1, 1, 1, 0],
                  [1, 1, 0, 1],
                  [0, 1, 1, 1],
                  [1, 0, 1, 1]], np.int8)
    y = np.array([1, 1, 1, 1], np.int8)
    return (X, y)


def test_default_params():
    clf = BooleanDecisionRuleClassifier()
    assert clf.max_iterations == -1
    assert clf.C == 10
    assert clf.p == 1
    assert clf.rmp_solver_params == ""
    assert clf.master_ip_solver_params == ""
    assert clf.subproblem_params == [""]
    assert clf.rmp_is_ip


def test_fit_predict(data):
    clf = BooleanDecisionRuleClassifier(max_iterations=3, C=10, p=1)
    clf.fit(data[0], data[1])
    assert hasattr(clf, 'is_fitted_')

    assert hasattr(clf, 'classes_')
    assert hasattr(clf, 'X_')
    assert hasattr(clf, 'y_')
    assert hasattr(clf, 'mp_optimal_')
    assert hasattr(clf, 'performed_iter_')
    assert hasattr(clf, 'num_col_added_sp_')

    assert_equal(clf.mp_optimal_, True)
    assert_equal(clf.performed_iter_, 2)
    assert_equal(clf.num_col_added_sp_[0], 1)

    X = data[0]
    y_pred = clf.predict(X)
    assert y_pred.shape == (X.shape[0],)
    assert_array_equal(y_pred, data[1])


def test_fit_predict_one_class(data_one_class):
    clf = BooleanDecisionRuleClassifier(max_iterations=3, C=10, p=1)
    with assert_raises(ValueError):
        clf.fit(data_one_class[0], data_one_class[1])


def test_heuristics(data):
    clf = BooleanDecisionRuleClassifierWithHeuristic(
        max_iterations=3, C=10, p=1)
    clf.fit(data[0], data[1])
    assert hasattr(clf, 'is_fitted_')

    assert hasattr(clf, 'classes_')
    assert hasattr(clf, 'X_')
    assert hasattr(clf, 'y_')
    assert hasattr(clf, 'mp_optimal_')
    assert hasattr(clf, 'performed_iter_')
    assert hasattr(clf, 'num_col_added_sp_')

    assert_equal(clf.mp_optimal_, True)
    assert_equal(clf.performed_iter_, 3)
    assert_equal(clf.num_col_added_sp_[0], 2)
    assert_equal(clf.num_col_added_sp_[1], 1)

    X = data[0]
    y_pred = clf.predict(X)
    assert y_pred.shape == (X.shape[0],)
    assert_array_equal(y_pred, data[1])


def test_mp_satisfies_clause():
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    assert_equal(master_problem.satisfies_clause(
        [0, 1, 1, 1, 0, 0, 0], [1, 2, 3]), 1)
    assert_equal(master_problem.satisfies_clause(
        [0, 1, 1, 1, 0, 0, 0], [2, 3]), 1)
    assert_equal(master_problem.satisfies_clause(
        [0, 1, 1, 1, 0, 0, 0], [0, 2, 3]), 0)
    assert_equal(master_problem.satisfies_clause(
        [0, 1, 1, 1, 0, 0, 0], [1, 2, 4]), 0)


def test_generate_lexicographic_clause():
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    assert_array_equal(master_problem.generate_lexicographic_clause(0), [])
    assert_array_equal(master_problem.generate_lexicographic_clause(5), [0, 2])
    assert_array_equal(master_problem.generate_lexicographic_clause(6), [1, 2])


def test_get_objective_coeff_mp(data):
    # Test how many zero examples satisfies the clause
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    master_problem.generate_mp(data[0], data[1])
    assert_equal(master_problem.get_objective_coeff_mp(clause=[0, 3]), 1)
    assert_equal(master_problem.get_objective_coeff_mp(clause=[2]), 2)
    assert_equal(master_problem.get_objective_coeff_mp(clause=[1, 2]), 1)
    assert_equal(master_problem.get_objective_coeff_mp(clause=[]), 2)
    assert_equal(master_problem.get_objective_coeff_mp(clause=[1, 3]), 1)
    assert_equal(master_problem.get_objective_coeff_mp(clause=[0, 1]), 0)


def test_get_clause_coeffs(data):
    # [positive examples] Coeff = 1 if example satisfies clause,
    # [complexity] Coeff = size of clause +1
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    assert_equal(master_problem.generated_, False)
    master_problem.generate_mp(data[0], data[1])
    assert_equal(master_problem.generated_, True)
    assert_array_equal(master_problem.get_clause_coeffs(
        clause=[0, 3]), [0, 1, 3])
    assert_array_equal(master_problem.get_clause_coeffs(clause=[2]), [1, 0, 2])
    assert_array_equal(master_problem.get_clause_coeffs(
        clause=[0, 1]), [1, 1, 3])


def test_duals(data):
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    master_problem.generate_mp(data[0], data[1])
    cc_dual, cs_duals = master_problem.solve_rmp()
    assert_equal(cc_dual, 0)
    assert_array_equal(cs_duals, [1, 1])


def test_add_column(data):
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    master_problem.generate_mp(data[0], data[1])
    clause = [0, 1]  # Clause that is only satisfied by positive examples.
    assert_equal(master_problem.add_column(clause), True)
    solved = master_problem.solve_ip()
    assert_equal(solved, True)
    assert_array_equal(master_problem.explanation, [[0, 1]])


def test_add_empty_column(data):
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    master_problem.generate_mp(data[0], data[1])
    clause = []
    assert_equal(master_problem.add_column(clause), False)


def test_regenerate_mp(data):
    # This test should cover the 'if' check at the start of 'generate_mp'
    # method.
    master_problem = BDRMasterProblem(
        C=10, p=1, optimization_problem_type='glop')
    assert_equal(master_problem.generated_, False)
    master_problem.generate_mp(data[0], data[1])
    assert_equal(master_problem.generated_, True)
    master_problem.generate_mp(data[0], data[1])
    assert_equal(master_problem.generated_, True)


def test_sp_generate_column(data):
    subproblem = BDRSubProblem(D=10, optimization_problem_type='cbc')
    assert_equal(subproblem.generated_, False)
    clauses = subproblem.generate_columns(
        data[0], data[1], dual_costs=(0, [1, 1]))
    assert_array_equal(clauses, [[0, 1]])
    assert_equal(subproblem.generated_, True)
    # Call generate again. Only objective should be updated.
    clauses = subproblem.generate_columns(
        data[0], data[1], dual_costs=(0, [0, 0]))
    assert_array_equal(clauses, [])


def test_heuristic_generate_column(data):
    subproblem = BDRHeuristic(K=1)
    clauses = subproblem.generate_columns(
        data[0], data[1], dual_costs=(0, [1, 1]))
    assert_array_equal(clauses, [[0], [1]])

    # Call generate again.
    clauses = subproblem.generate_columns(
        data[0], data[1], dual_costs=(0, [0.5, 0.5]))
    assert_array_equal(clauses, [])


def test_heuristic_generate_column2(data):
    subproblem = BDRHeuristic(K=2)
    clauses = subproblem.generate_columns(
        data[0], data[1], dual_costs=(0, [1, 1]))
    assert_array_equal(clauses, [[0], [1], [0, 1]])



"""
Base Column Genaration Classifier. One needs to extend this for implementing
column generation based classifiers.
"""

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y
from sklearn.utils.multiclass import unique_labels
from sklearn.preprocessing import LabelEncoder


class BaseMasterProblem():
    """ Base class for master problem. One needs to extend this for using with
    ColGenClassifier.
    """

    def __init__(self):
        pass

    def generate_mp(self, X, y):
        """ Generates the master problem model (RMP) and initializes the primal
        and dual solutions.
        Parameters
        ----------
        X : ndarray, shape (n_samples, n_features)
            The input.
        y : ndarray, shape (n_samples,)
            The labels.
        """
        pass

    def add_column(self, column):
        """ Adds the given column to the master problem model.
        Parameters
        ----------
        column : object,
            The column to be added to the RMP.
        """
        pass

    def solve_rmp(self, solver_params=''):
        """ Solves the RMP with given solver params.
        Returns the dual costs.
        Parameters
        ----------
        solver_params : string, default='',
            The solver parameters for solving the RMP.
        """
        pass

    def solve_ip(self, solver_params=''):
        """ Solves the integer RMP with given solver params.
        Returns false if the problem is not solved.
        Parameters
        ----------
        solver_params : string, default='',
            The solver parameters for solving the integer RMP.
        """
        pass


class BaseSubproblem():
    """ Base class for subproblem. One needs to extend this for using with
    ColGenClassifier.
    """

    def generate_columns(self, X, y, dual_costs, params):
        """ Generates the new columns to be added to the RMP.
        """
        pass


class ColGenClassifier(ClassifierMixin, BaseEstimator):
    """ Base Column Genaration Classifier. One needs to extend this for
    implementing column generation based classifiers.

    Parameters
    ----------
    max_iterations : int, default=-1
        Maximum column generation iterations. Negative values removes the
        iteration limit and the problem is solved till optimality.
    master_problem : Instance of BaseMasterProblem, default = BaseSubproblem()
    subproblem : Instance of BaseSubproblem, default = BaseSubproblem()
    rmp_is_ip : boolean, default = False
        True if the master problem has integer variables.
    rmp_solver_params: string, default = "",
        Solver parameters for solving restricted master problem (rmp).
    master_ip_solver_params: string, default = "",
        Solver parameters for solving the integer master problem.
    subproblem_params: list of string, default = [""],
        Parameters for solving the subproblem.

    Attributes
    ----------
    X_ : ndarray, shape (n_samples, n_features)
        The input passed during :meth:`fit`.
    y_ : ndarray, shape (n_samples,)
        The labels passed during :meth:`fit`.
    classes_ : ndarray, shape (n_classes,)
        The classes seen at :meth:`fit`.
    performed_iter_ : int
        Total number of iterations performed in column generation loop.
    mp_optimal: bool
        Set to true if none of the subproblems could generate any column in an
        iteration.
    num_col_added_sp_: list of ints (size = number of subproblems)
        Count of number of columns added to the master problem by each
        subproblem.
    """

    def __init__(self, max_iterations=-1,
                 master_problem=BaseMasterProblem(),
                 subproblems=[BaseSubproblem()],
                 rmp_is_ip=False,
                 rmp_solver_params="",
                 master_ip_solver_params="",
                 subproblem_params=[""]):
        self.max_iterations = max_iterations
        self.master_problem = master_problem
        self.subproblems = subproblems
        self.rmp_is_ip = rmp_is_ip
        self.rmp_solver_params = rmp_solver_params
        self.master_ip_solver_params = master_ip_solver_params
        self.subproblem_params = subproblem_params
        assert len(subproblems) == len(subproblem_params)

    def fit(self, X, y):
        """Runs the column generation loop.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.

        Returns
        -------
        self : object
            Returns self.
        """
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y

        # Convert the y into integer class lables.
        self.label_encoder_ = LabelEncoder()
        self.processed_y_ = self.label_encoder_.fit_transform(y)
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)
        if len(self.classes_) <= 1:
            raise ValueError(
                "Classifier can't train when only one class is present.")

        # Initiate the master and subproblems
        self.master_problem.generate_mp(X, self.processed_y_)

        self.performed_iter_ = 0
        self.mp_optimal_ = False
        self.num_col_added_sp_ = [0]*len(self.subproblems)

        for iter in range(self.max_iterations):
            self.performed_iter_ += 1
            dual_costs = self.master_problem.solve_rmp(self.rmp_solver_params)

            rmp_updated = False
            sp_ind = 0
            for sp_ind in range(len(self.subproblems)):
                generated_columns = self.subproblems[sp_ind].generate_columns(
                    X, self.processed_y_, dual_costs,
                    self.subproblem_params[sp_ind])

                rmp_updated = False
                for column in generated_columns:
                    col_added = self.master_problem.add_column(column)
                    self.num_col_added_sp_[sp_ind] += 1 if col_added else 0
                    rmp_updated = rmp_updated or col_added
                if rmp_updated:
                    break

            if not rmp_updated:
                print("RMP not updated. exiting the loop.")
                self.mp_optimal_ = True
                break

        if self.rmp_is_ip:
            solved = self.master_problem.solve_ip(self.master_ip_solver_params)
            assert solved, "RMP integer program couldn't be solved."

        self.is_fitted_ = True
        # Return the classifier
        return self

    def predict(self, X):
        """ Predicts the class based on the solution of master problem. This
        method is abstract.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample.
        """
        pass



#! /usr/bin/env python
"""A template for scikit-learn compatible packages."""

import codecs

from setuptools import find_packages, setup
from col_gen_estimator._version import __version__

DISTNAME = 'ColGenEstimator'
DESCRIPTION = 'Classifiers using column generation'
with codecs.open('README.rst', encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
MAINTAINER = 'K. Patel'
MAINTAINER_EMAIL = 'krunalkachadia@gmail.com'
URL = 'https://github.com/krooonal/col_gen_estimator'
LICENSE = 'new BSD'
DOWNLOAD_URL = 'https://github.com/krooonal/col_gen_estimator'
VERSION = __version__
INSTALL_REQUIRES = ['numpy', 'scipy', 'scikit-learn', 'ortools', 'bitarray']
CLASSIFIERS = ['Intended Audience :: Science/Research',
               'Intended Audience :: Developers',
               'License :: OSI Approved',
               'Programming Language :: Python',
               'Topic :: Software Development',
               'Topic :: Scientific/Engineering',
               'Operating System :: Microsoft :: Windows',
               'Operating System :: POSIX',
               'Operating System :: Unix',
               'Operating System :: MacOS',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7']
EXTRAS_REQUIRE = {
    'tests': [
        'pytest',
        'pytest-cov'],
    'docs': [
        'sphinx',
        'sphinx-gallery',
        'sphinx_rtd_theme',
        'numpydoc',
        'matplotlib'
    ]
}

setup(name=DISTNAME,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      license=LICENSE,
      url=URL,
      version=VERSION,
      download_url=DOWNLOAD_URL,
      long_description=LONG_DESCRIPTION,
      zip_safe=False,  # the package can run out of an .egg file
      classifiers=CLASSIFIERS,
      packages=find_packages(),
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE)
